#!/usr/bin/env python

#***************************************************************************
#                        libcdemu.py  -  description
#                             -------------------
#    copyright            : (C) 2003 by Robert Penz
#    email                : robert.penz@outertech.com
#***************************************************************************

#***************************************************************************
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU General Public License as published by  *
#*   the Free Software Foundation; either version 2 of the License, or     *
#*   (at your option) any later version.                                   *
#*                                                                         *
#***************************************************************************

import sys
import os
import re
import string
import array
import struct
import fcntl
import stat

# this const is used by get_device_from_drive_number
# which adds the drive number at the end of this string
device_path="/dev/cdemu"

CUE_TRACK   = 0
CUE_MODE    = 1
CUE_INDEXES = 2

ENTRY_MODE  = 0
ENTRY_MIN   = 1
ENTRY_SEC   = 2
ENTRY_FRAME = 3
ENTRY_STARTSECTOR = 4
ENTRY_STARTOFFSET = 5

CDEMU_LOADCD = 0x5400   # load the virtual cd
CDEMU_UNLOADCD = 0x5401 # unload the virtual cd

UNKNOWN = -1
AUDIO = 0
MODE1 = 1
MODE2 = 2
AUDIO_2448 = 10
MODE1_2448 = 20
MODE1_2352 = 30
MODE1_2048 = 40
MODE2_2352 = 50
MODE2_2336 = 60
LEADOUT = 100

# Some generally useful CD-ROM information -- copy and past from include/linux/cdrom.h
CD_MINS            =   74 #/* max. minutes per CD, not really a limit */
CD_SECS            =   60 #/* seconds per minute */
CD_FRAMES          =   75 #/* frames per second */
CD_SYNC_SIZE       =   12 #/* 12 sync bytes per raw data frame */
CD_MSF_OFFSET      =  150 #/* MSF numbering offset of first frame */
CD_CHUNK_SIZE      =   24 #/* lowest-level "data bytes piece" */
CD_NUM_OF_CHUNKS   =   98 #/* chunks per frame */
CD_FRAMESIZE_SUB   =   96 #/* subchannel data "frame" size */
CD_HEAD_SIZE       =    4 #/* header (address) bytes per raw data frame */
CD_SUBHEAD_SIZE    =    8 #/* subheader bytes per raw XA data frame */
CD_EDC_SIZE        =    4 #/* bytes EDC per most raw data frame types */
CD_ZERO_SIZE       =    8 #/* bytes zero per yellow book mode 1 frame */
CD_ECC_SIZE        =  276 #/* bytes ECC per most raw data frame types */
CD_FRAMESIZE       = 2048 #/* bytes per frame, "cooked" mode */
CD_FRAMESIZE_RAW   = 2352 #/* bytes per frame, "raw" mode */
CD_FRAMESIZE_RAWER = 2646 #/* The maximum possible returned bytes */
# most drives don't deliver everything:
CD_FRAMESIZE_RAW1  = (CD_FRAMESIZE_RAW-CD_SYNC_SIZE)              # /*2340*/
CD_FRAMESIZE_RAW0  = (CD_FRAMESIZE_RAW-CD_SYNC_SIZE-CD_HEAD_SIZE) # /*2336*/

#typedef struct track
#{
#   unsigned short mode;
#   unsigned short minute;
#   unsigned short second;
#   unsigned short frame;
#
#   /* (min*60 + sec) * 75 + fps   */
#
#   unsigned long start_sector;
#
#   /* = the sizes in bytes off all tracks bevor this one */
#   /* its needed if there are mode1 tracks befor the mpeg tracks */
#   unsigned long start_offset;
#
#} tTrack;
track_struct = "HHHHLL"
# only 47 possible as http://mail.python.org/pipermail/python-checkins/2003-March/034383.html
tracks_size = struct.calcsize(track_struct) * 47

#struct virtualcd_info {
#   tTrack          tracks[100];
#   unsigned int    ffd;
#   int             num_tracks; /* number of tracks on the cd */
#  char            comment[256];
#};

cdemu_struct = "%dsIi255sx" % tracks_size

# default verbose is off
# 1 is on and 2 (currently used even more)
verbose = 0

class CDEmu_Error(StandardError):
    "our cdemu custom error class"

def read_binary(fmt, what, offset):
    "helper function for reading binary data into variables"
    if (fmt == "B") or (fmt == "<B") or (fmt == ">B") :
        return struct.unpack(fmt, what[offset])[0]
    elif (fmt == "H") or (fmt == "<H") or (fmt == ">H") :
        return struct.unpack(fmt, what[offset:offset+2])[0]
    elif (fmt == "h") or (fmt == "<h") or (fmt == ">h") :
        return struct.unpack(fmt, what[offset:offset+2])[0]
    elif (fmt == "L") or (fmt == "<L") or (fmt == ">L") :
        return struct.unpack(fmt, what[offset:offset+4])[0]
    elif (fmt == "l") or (fmt == "<l") or (fmt == ">l") :
        return struct.unpack(fmt, what[offset:offset+4])[0]
    elif (fmt == "Q") or (fmt == "<Q") or (fmt == ">Q") :
        return struct.unpack(fmt, what[offset:offset+8])[0]
    else:
        raise CDEmu_Error, "Error in reading binary data"

def text2number(text):
    "helper function for reading a number from text into a variable"
    if text[0:2] == "0x":
        return int(text, 16)
    else:
        return int(text)

def parse_iso(iso_filename):
    "returns the supposable bin_filename and raw entries, if something went \
    wrong it throws an exception"
    # it's an ISO9660 file -- so emulate a 1 track, 2048 sector-size 
    # BIN/CUE file
    if verbose: print "ISO9660 file: " , iso_filename
    entries = [ [ "01", "MODE1/2048", [ ["01", "00:00:00"] ] ] ]

    return iso_filename, entries

def parse_mds_track(mds_content, track_offset):
    mds_extrablock_size = 0x08
    mds_footer_size = 0x16

    mode = read_binary("H", mds_content, track_offset+0x0000)

    flags = read_binary("B", mds_content, track_offset+0x0002)
    ctrl = flags & 0xf
    adr = (flags >> 4) & 0xf

    tno = read_binary("B", mds_content, track_offset+0x0003)
    point = read_binary("B", mds_content, track_offset+0x0004)
    min = read_binary("B", mds_content, track_offset+0x0005)
    sec = read_binary("B", mds_content, track_offset+0x0006)
    frac = read_binary("B", mds_content, track_offset+0x0007)
    zero = read_binary("B", mds_content, track_offset+0x0008)
    pmin = read_binary("B", mds_content, track_offset+0x0009)
    psec = read_binary("B", mds_content, track_offset+0x000a)
    pfrac = read_binary("B", mds_content, track_offset+0x000b)
    extra_offset = read_binary("L", mds_content, track_offset+0x000c)
    sectorsize = read_binary("H", mds_content, track_offset+0x0010)
    lba = read_binary("L", mds_content, track_offset+0x0024)
    offset = read_binary("Q", mds_content, track_offset+0x0028)

    if point <= 0x99:
	pregap = read_binary("l", mds_content, extra_offset+0x0000)
	sectors = read_binary("l", mds_content, extra_offset+0x0004)
    else:
    	pregap = 0
	sectors = 0

    if verbose:
	    print "    point: %2x, mode: %3x, control: %x, adr: %i, MSF: %02i:%02i:%02i, P-MSF: %02i:%02i:%02i, sector size: %4i, lba: %i, sectors: %i, offset: %i" % \
	      (point, mode, ctrl, adr, min, sec, frac, pmin, psec, pfrac, sectorsize, lba, sectors, offset)

    return [point, mode, sectorsize, lba, sectors, offset]

def parse_mds_session(mds_content, session_offset):
    pregap_corr = read_binary("l", mds_content, session_offset + 0x0000)
    num_sectors = read_binary("L", mds_content, session_offset + 0x0004)
    num_tracks = read_binary("B", mds_content, session_offset + 0x000a)
    num_leadins = read_binary("B", mds_content, session_offset + 0x000b)
    first_track = read_binary("B", mds_content, session_offset + 0x000c)
    last_track = read_binary("B", mds_content, session_offset + 0x000e)
    tracks_offset = read_binary("L", mds_content, session_offset + 0x0014)

    if verbose:
	print "  pregap_corr: %i, num_sectors: %i, num_tracks: %i, num_leadins: %i, first_track: %i, last_track: %i, tracks_offset: %i" % \
		(pregap_corr, num_sectors, num_tracks, num_leadins, first_track, last_track, tracks_offset)

    track_struct_size = 0x50

    session = []

    # read datablocks and extrablocks
    for i in range(0, num_tracks):
	track = parse_mds_track(mds_content, tracks_offset + track_struct_size * i)
	# writing data to entries
	if track[0] < 0xa0:
	    session.append(track)

    return session

def parse_mds(mds_filename):
    "returns the supposable bin_filename and raw entries, if something went \
    wrong it throws an exception"
    # read the file
    try:
        mds_file = open (mds_filename,"rb")
    except:
        raise CDEmu_Error, "Could not open the mds-file <%s>" % mds_filename

    mds_content = mds_file.read()
    mds_size = len(mds_content)
    
    if mds_content[0x0000:0x0010] != "MEDIA DESCRIPTOR":
        raise CDEmu_Error, "Invalid data in <%s>. It is not an MDF/MDS file." % mds_filename

    mds_version = read_binary("B", mds_content, 0x0010)
    mds_revision = read_binary("B", mds_content, 0x0011)
    mds_sessions = read_binary("H", mds_content, 0x0014)
    mds_sessions_offset = read_binary("L", mds_content, 0x0050)
    if verbose:
        print "MDS/MDF version: %i.%i" % (mds_version, mds_revision)
        print "sessions: %i" % (mds_sessions)

    # set name of binary cd-image
    bin_filename = os.path.splitext(os.path.basename(mds_filename))[0] + ".mdf";
    if verbose:
        print "binfile: %s" % bin_filename

    mds_session_size = 0x18

    entries = []

    for session in range(0, mds_sessions):
	print "session %i:" % (session)
	entries = parse_mds_session(mds_content, mds_sessions_offset + mds_session_size * session)
	break

    entries.sort() # python is sexy.

    return bin_filename, entries

def parse_nrg(nrg_filename):
    "returns the supposable bin_filename and raw entries, if something went \
    wrong it throws an exception"
    # read the file
    try:
        nrg_file = open (nrg_filename,"rb")
    except:
        raise CDEmu_Error, "Could not open the nrg-file <%s>" % nrg_filename

    # last 8 or 12 bytes gives us important information, like the signature
    # and where to find the rest of the image information.
    nrg_file.seek(0, 2)
    nrg_filesize = nrg_file.tell()

    nrg_file.seek(-12, 2)
    nrg_signature = nrg_file.read(4)
    if nrg_signature == "NER5": # Nero 5.5.x or newer
        if verbose: print "File format: NER5."
        nrg_trailer_offset = read_binary(">Q", nrg_file.read(8), 0)
        oldformat = 0 # Offsets are 64 bit.
    else:
        nrg_file.seek(-8, 2)
        nrg_signature = nrg_file.read(4)
        if nrg_signature == "NERO": # Nero older than 5.5.x
            if verbose: print "File format: NERO."
            nrg_trailer_offset = read_binary(">L", nrg_file.read(4), 0)
            oldformat = 1 # Offsets are 32 bit.
        else:
            raise CDEmu_Error, "Invalid data in <%s>. It is not an NRG file." % nrg_filename

    nrg_file.seek(nrg_trailer_offset, 0)
    nrg_content = nrg_file.read(nrg_filesize - nrg_trailer_offset)
    
    # find relative offsets for different blocks
    offs = 0
    # CUEX is 64-bit variant of CUES, same goes for DAOX vs. DAOI and ETN2 vs. ETNF.
    # DAO is DiscAtOnce, ETN is TrackAtOnce. SINF is Session Information, and MTYP is Media Type.
    # CDTX is CD-Text. END! is a marker for the last block.
    block_ids = ("CUEX", "CUES", "DAOX", "DAOI", "CDTX", "ETNF", "ETN2", "SINF", "MTYP", "END!")
    block = {}
    while offs < nrg_filesize - nrg_trailer_offset - 12:
        for block_id in block_ids:
            if(block_id == nrg_content[offs:offs+4]):
                size = read_binary(">L", nrg_content, offs+4)
                block[block_id] = (offs, size)
                if verbose: print "%s [0x%x:0x%x]" % (block_id, offs, size)
                offs += size+8
                break
        else:
            raise CDEmu_Error, "Error in parsing Nero CD-Image."

    # DiscAtOnce or TrackAtOnce?
    if oldformat:
        if block.has_key("DAOI") and block.has_key("CUES"): is_dao = 1
        elif block.has_key("SINF") and block.has_key("ETNF"): is_dao = 0
        else: raise CDEmu_Error, "Error in checking for DAO / TAO."
    else:
        if block.has_key("DAOX") and block.has_key("CUEX"): is_dao = 1
        elif block.has_key("SINF") and block.has_key("ETN2"): is_dao = 0
        else: raise CDEmu_Error, "Error in checking for DAO / TAO."
    if verbose:
        if is_dao: print "CD-Image is Disc-At-Once."
        else: print "CD-Image is Track-At-Once."
            
    # Got CD-Text? Parse it.
    if block.has_key("CDTX"):
        if verbose: print "CD-TEXT:"
        entry = 0
        cdtext_song = ""
        cdtext_artist = ""
        cdtext_compilation = ""
        while entry*18+8 < block["CDTX"][1]: 
            cdtext_1 = read_binary(">H", nrg_content, block["CDTX"][0]+8+entry*18+0)
            cdtext_2 = read_binary(">H", nrg_content, block["CDTX"][0]+8+entry*18+2)
            cdtext_3 = nrg_content[block["CDTX"][0]+8+entry*18+4:block["CDTX"][0]+8+entry*18+16]
            cdtext_4 = read_binary(">H", nrg_content, block["CDTX"][0]+8+entry*18+0)
            cdtext_entry = cdtext_2 >> 8
            cdtext_type = cdtext_1 >> 8
            if (entry == 0 and cdtext_entry == 0) or (cdtext_entry != 0):
                if cdtext_type == 0x80: cdtext_song += cdtext_3
                elif cdtext_type == 0x81: cdtext_artist += cdtext_3
                elif cdtext_type == 0x8f: cdtext_compilation += cdtext_3
                else: print "NOTE: Unrecognized cd-text type found while parsing Nero CD-Image."
            #if verbose:
            #    print "[%3i] 0x%04x 0x%04x %s 0x%04x" % (entry, cdtext_1, cdtext_2, cdtext_3, cdtext_4)
            entry+=1
        cdtext_song = string.split(cdtext_song, "\0")
        cdtext_artist = string.split(cdtext_artist, "\0")
        while len(cdtext_song[-1]) == 0: del cdtext_song[-1]
        while len(cdtext_artist[-1]) == 0: del cdtext_artist[-1]
        if verbose:
            for entry in range(0, len(cdtext_song)):
                print "Artist: %s   Song: %s" % (cdtext_artist[entry], cdtext_song[entry])

    # making sure table is empty
    entries = []
    
    if verbose: print "TRACK ENTRIES:"
    if is_dao: # Read DAO information.
        entry = 1 # entry 0 is block-header entry so we start on entry 1
        while 1:
            if(oldformat):
                cue_offset = block["CUES"][0]
                dao_offset = block["DAOI"][0]
            else:
                cue_offset = block["CUEX"][0]
                dao_offset = block["DAOX"][0]
    
            # tracktype. 0x40 = datatrack, 0x01 = copyright?
            tracktype = read_binary(">B", nrg_content, cue_offset+8+entry*16+0)
            track_raw = read_binary(">B", nrg_content, cue_offset+8+entry*16+1)
            track = (track_raw >> 4) * 10 + (track_raw & 0x0f)
            # addr_ctrl. addresstype MSF or LBA in lower 4 bits, Q-channel control field in upper 4 bits.
            addr_ctrl = read_binary(">B", nrg_content, cue_offset+8+entry*16+2)
            start_sector = read_binary(">l", nrg_content, cue_offset+8+entry*16+4)
            nexttrack_raw = read_binary(">B", nrg_content, cue_offset+8+entry*16+9)
            nexttrack = (nexttrack_raw >> 4) * 10 + (nexttrack_raw & 0x0f)
            end_sector = read_binary(">l", nrg_content, cue_offset+8+entry*16+12)
            
            isrc = nrg_content[dao_offset-12+entry*42+0:dao_offset-12+entry*42+12]
            sector_size = read_binary(">H", nrg_content, dao_offset-12+entry*42+12)
            mode = read_binary(">B", nrg_content, dao_offset-12+entry*42+14)
    
            if(oldformat):
                start_offset = read_binary(">L", nrg_content, dao_offset-12+entry*42+22)
                end_offset = read_binary(">L", nrg_content, dao_offset-12+entry*42+26)
            else:
                start_offset = read_binary(">Q", nrg_content, dao_offset-12+entry*42+26)
                end_offset = read_binary(">Q", nrg_content, dao_offset-12+entry*42+34)
    
            if verbose:
                print "[%3i] track: %3i, next track: %3i, start sector: %8i, end sector: %8i\n      sector size: %i, mode: %i, start: 0x%08x, end: 0x%08x." % \
                (entry, track, nexttrack, start_sector, end_sector, sector_size, mode, start_offset, end_offset)
                print "      tracktype: 0x%x, addr_ctrl: 0x%x, isrc: %s." % (tracktype, addr_ctrl, isrc)
            entries.append([track, track_raw, nexttrack, nexttrack_raw, start_sector, end_sector, sector_size, mode, start_offset, end_offset])
            entry += 1
            if nexttrack_raw == 0xAA: break # lead-out
    else: # Read TAO information.
        tracks_in_session = read_binary(">L", nrg_content, block["SINF"][0]+8+0)
        if verbose: print "Tracks in Session: %i." % (tracks_in_session)
                        
        for entry in range(0, tracks_in_session):
            if oldformat:
                offset = read_binary(">L", nrg_content, block["ETNF"][0]+8+entry*20+0)
                size = read_binary(">L", nrg_content, block["ETNF"][0]+8+entry*20+4)
                mode = read_binary(">B", nrg_content, block["ETNF"][0]+8+entry*20+11)
                sector = read_binary(">L", nrg_content, block["ETNF"][0]+8+entry*20+12)
            else:
                offset = read_binary(">Q", nrg_content, block["ETN2"][0]+8+entry*32+0)
                size = read_binary(">Q", nrg_content, block["ETN2"][0]+8+entry*32+8)
                mode = read_binary(">B", nrg_content, block["ETN2"][0]+8+entry*32+19)
                sector = read_binary(">L", nrg_content, block["ETN2"][0]+8+entry*32+20)
            if verbose:
                print "[%3i] offset: 0x%08x, size: 0x%08x, sector: %i, mode: %x" % (entry+1, offset, size, sector, mode)
            entries.append([entry+1, entry+1, entry+2, entry+2, sector, 0, 0, mode, offset, offset+size])

    entries.sort() # python is sexy.

    # set name of binary cd-image
    bin_filename = nrg_filename;
    if verbose:
        print "binfile: %s" % bin_filename

    return bin_filename, entries

def parse_cue(cue_filename):
    "returns the supposable bin_filename and raw entries, if something went \
    wrong it throws an exception"
    # don't require an excact match .. so forget the ^ and $  ;-)
    regex_header = re.compile(r'FILE\s*"?(?P<bin_filename>.+?)"?\s*BINARY', re.IGNORECASE)
    regex_track = re.compile(r'TRACK\s*(?P<track>\d+)\s*(?P<mode>\S+)', re.IGNORECASE)
    regex_index = re.compile(r'INDEX\s*(?P<index>\d+)\s*(?P<msf>\d{1,2}\:\d{1,2}\:\d{1,2})', re.IGNORECASE)
    regex_ignore = re.compile(r'PREGAP\s|FLAGS\s', re.IGNORECASE)

    # read the file
    try:
        input = open (cue_filename,"r").readlines()
    except:
        raise CDEmu_Error, "Could not open the cue-file <%s>" % cue_filename
    
    # remove empty lines
    for i in range(len(input) - 1, 0):
        if not input[i].strip().rstrip('\n'):
            del(input[i])
    
    # the first line is the header .. match it
    found = regex_header.search(input[0])
    if found:
        if verbose: print "binfile: %s" % found.group("bin_filename")
        bin_filename = found.group("bin_filename")
    else:
        raise CDEmu_Error, "Invalid header of the cue-file <%s>" % cue_filename

    # remove the header
    del input[0]
    entries = []
 
    # look at the rest
    for line in input:
        found = regex_track.search(line)
        if found:
            if verbose: print "track: %s mode: %s" % (found.group("track"), found.group("mode"))
            entries.append([found.group("track"), found.group("mode"),[]])
            continue
    
        found = regex_index.search(line)
        if found:
            if verbose: print "index: %s msf: %s" % (found.group("index"), found.group("msf"))
            if entries != []:
                # insert my index data to the last track
                entries[-1][CUE_INDEXES].append([found.group("index"), found.group("msf")])
            else:
                if verbose: print "index before the first track --> ignoring"
            continue

        if regex_ignore.search(line):
            continue

        # we found a line we couldn't match ... thats an error
        print "Couldn't match: <%s>  (spaces striped)" % string.strip(line)
        
    return bin_filename, entries

def parse_ccd(ccd_filename):
    "returns the supposable bin_filename and raw entries, if something went \
    wrong it throws an exception"
    # match known data blocks
    regex_clonecd = re.compile(r'\[CloneCD\]', re.IGNORECASE)
    regex_disc = re.compile(r'\[Disc\]', re.IGNORECASE)
    regex_session = re.compile(r'\[Session\s(?P<session>\d+)\]', re.IGNORECASE)
    regex_entry = re.compile(r'\[Entry\s(?P<entry>\d+)\]', re.IGNORECASE)
    regex_track = re.compile(r'\[TRACK\s(?P<track>\d+)\]', re.IGNORECASE)

    regex_cfgpair = re.compile(r'^(?P<key>[A-Za-z0-9\s^=]*)=(?P<content>.*)', re.IGNORECASE)
    regex_blocks = (("clonecd", regex_clonecd, 0), ("disc", regex_disc, 0), ("session", regex_session, 1),
                    ("entry", regex_entry, 1), ("track", regex_track, 1))
        
    # read the file
    try:
        input = open (ccd_filename,"r").readlines()
    except:
        raise CDEmu_Error, "Could not open the ccd-file <%s>" % ccd_filename
    
    # set name of binary cd-image
    bin_filename = os.path.splitext(os.path.basename(ccd_filename))[0] + ".img";
    if verbose:
        print "binfile: %s" % bin_filename

    # making sure dictionaries is empty
    block = {}
    num = {}

    # look at the rest
    i=-1
    while i < (len(input) - 1):
        i+=1
        line = input[i]
        for regex_block in regex_blocks:
            block_name = regex_block[0]
            regex = regex_block[1]
            return_list_entry = regex_block[2]
            
            found = regex.search(line)
            if found:
                if return_list_entry:
                    num[block_name] = text2number(found.group(block_name))
                temp_dict = {}
                while (i < (len(input) - 1) and input[i].strip().rstrip('\n')):
                    i+=1
                    line = input[i]
                    found = regex_cfgpair.search(line)
                    if found: temp_dict[found.group("key").lower()] = text2number(found.group("content"))
                    else: break
                if return_list_entry:
                    if block.has_key(block_name): block[block_name].append(temp_dict)
                    else: block[block_name] = [temp_dict]
                else:
                    block[block_name] = temp_dict
                i-=1;
                break
                
    if verbose:
        print "CLONECD: "; print block["clonecd"]
        print "DISC: "; print block["disc"]
        print "SESSION: "; print block["session"]
        print "ENTRY: "; print block["entry"]
        print "TRACK: "; print block["track"]

    return bin_filename, block

def find_bin(supposable_bin_filename, cue_filename):
    "returns the fullpath of the bin file if found - the cue_filename must be with full path"

    plain_bin = os.path.abspath(supposable_bin_filename)
    cue_path = os.path.dirname(cue_filename)
    bin_basename = os.path.basename(supposable_bin_filename)
    bin_in_cue_path = os.path.join(cue_path, bin_basename)
    bin_in_bin_path = os.path.abspath(bin_basename)
    bin_no_extension = os.path.join(cue_path, os.path.splitext(bin_basename)[0])

    filenames_2_try = [ plain_bin, bin_in_cue_path, bin_in_cue_path.lower(), bin_in_cue_path.upper(),
                        bin_in_bin_path, bin_in_bin_path.lower(), bin_in_bin_path.upper(),
                        bin_no_extension + ".bin", bin_no_extension + ".img", bin_no_extension + ".mdf",
                        bin_no_extension + ".iso", bin_no_extension + ".nrg" ]

    for filename in filenames_2_try:
        if verbose: print "Trying to open <%s>" % filename
        if os.path.isfile(filename): return filename
  
    raise CDEmu_Error, "Giving up, tried lots of different places, but couldn't find the bin-file <%s>" % supposable_bin_filename

def get_mode_cue(mode):
    "converts the string to our internal mode consts"
    if mode.upper() == "AUDIO": return AUDIO
    if mode.upper() == "MODE1/2352": return MODE1_2352
    if mode.upper() == "MODE1/2048": return MODE1_2048
    if mode.upper() == "MODE2/2352": return MODE2_2352
    if mode.upper() == "MODE2/2336": return MODE2_2336
  
    raise CDEmu_Error, "Unkown mode <%s>" % mode

def get_mode_ccd(mode):
    "converts the string to our internal mode consts"
    # for clonecd images sector size is always 2352.
    if mode == 0: return AUDIO
    if mode == 1: return MODE1_2352
    if mode == 2: return MODE2_2352
  
    raise CDEmu_Error, "Unknown mode number: %i" % (mode)

def get_mode_mds(mode, sectsize):
    "converts the data to our internal mode consts"
    mode &= 0xff
    if mode == 0xa9:
        if sectsize == 2448: return AUDIO_2448
        else: return AUDIO
    if mode == 0xaa:
        if sectsize == 2448: return MODE1_2448
        elif sectsize == 2352: return MODE1_2352
        elif sectsize == 2048: return MODE1_2048
        else: return MODE1
    if (mode == 0xab) or (mode == 0xac) or (mode == 0xad):
        if sectsize == 2352: return MODE2_2352
        elif sectsize == 2336: return MODE2_2336
        else: return MODE2
  
    raise CDEmu_Error, "Unknown mode number: %i" % (mode)

def get_mode_nrg(mode, sectsize):
    "converts the data to our internal mode consts"
    if mode == 0x07: return AUDIO
    if (mode == 0x00) or (mode == 0x05):
        if sectsize == 2352: return MODE1_2352
        elif sectsize == 2048: return MODE1_2048
        else: return MODE1
    if (mode == 0x01) or (mode == 0x06):
        if sectsize == 2352: return MODE2_2352
        elif sectsize == 2336: return MODE2_2336
        else: return MODE2
  
    raise CDEmu_Error, "Unknown mode number: %i" % (mode)

def mode_2_sectorsize(mode):
    if mode == AUDIO: return CD_FRAMESIZE_RAW
    if mode == AUDIO_2448: return CD_FRAMESIZE_RAW+CD_FRAMESIZE_SUB
    if mode == MODE1_2448: return CD_FRAMESIZE_RAW+CD_FRAMESIZE_SUB
    if mode == MODE1_2352: return CD_FRAMESIZE_RAW
    if mode == MODE1_2048: return CD_FRAMESIZE
    if mode == MODE2_2352: return CD_FRAMESIZE_RAW
    if mode == MODE2_2336: return CD_FRAMESIZE_RAW0

    raise CDEmu_Error, "Unknown mode number: %i" % (mode)

def get_msf(s):
    "converts 00:00:00 to 3 integers, take the 01 index if there is a choise - min, sec, frame"
    if len(s) > 1:
        for index in s:
            if int(index[0]) == 1:
                msf = index[1]
    else:
        msf = s[-1][1]
    min, sec, frame = msf.split(":")
    return int(min), int(sec), int(frame)

def msf_2_sector(min, sec, frame):
    "converts the msf to sector"
    return frame + (sec + min * 60 ) * 75 - CD_MSF_OFFSET;

def sector_2_msf(sector):
    "and the other way round - min,sec,frame"
    sector += CD_MSF_OFFSET
    sector, frame = divmod(sector, 75)
    sector, sec = divmod(sector, 60)
    return sector, sec, frame

def extract_track_data_cue(raw_entries, bin_filename):
    " generates out of the raw data, we got from the cue parser, useful data \
    we can use for our struct function"
    # first build an list we can calc with
  
    entries = []
  
    for i in range(0,len(raw_entries)):
        # mode / minute / second / frame
        mode = get_mode_cue(raw_entries[i][CUE_MODE])
        min, sec, frame = get_msf(raw_entries[i][CUE_INDEXES])
	# CUE's MSF should be offset by 00:02:00
        start_sector = msf_2_sector(min, sec, frame) + CD_MSF_OFFSET
        min, sec, frame = sector_2_msf(start_sector)
        
        if i == 0:
            start_offset = 0
        else:
            start_offset = entries[-1][ENTRY_STARTOFFSET] \
                + (start_sector -  entries[-1][ENTRY_STARTSECTOR]) * mode_2_sectorsize(entries[-1][ENTRY_MODE])

        entries.append([mode, min, sec, frame, start_sector, start_offset])

    # add the leadout
    mode = LEADOUT
    start_offset = os.path.getsize(bin_filename)
    start_sector = (start_offset - entries[-1][ENTRY_STARTOFFSET]) / mode_2_sectorsize(entries[-1][ENTRY_MODE]) + entries[-1][ENTRY_STARTSECTOR]
    min,sec,frame = sector_2_msf(start_sector)
    entries.append([mode, min, sec, frame, start_sector, start_offset])

    return entries

def extract_track_data_ccd(block, bin_filename):
    " generates out of the raw data, we got from the ccd parser, useful data \
    we can use for our struct function"
    # first build an list we can calc with
  
    entries = []
    skipped = 0
  
    for i in range(0,len(block["entry"])):
        # mode / minute / second / frame
        if block["entry"][i]["point"] >= 0xA0:
            skipped += 1
            continue
        mode = get_mode_ccd(block["track"][i-skipped]["mode"])
        start_sector = block["entry"][i]["plba"]
        min, sec, frame = sector_2_msf(start_sector)
        start_offset = start_sector * mode_2_sectorsize(mode)

        entries.append([mode, min, sec, frame, start_sector, start_offset])

    # add the leadout
    mode = LEADOUT
    start_offset = os.path.getsize(bin_filename)
    start_sector = (start_offset - entries[-1][ENTRY_STARTOFFSET]) / mode_2_sectorsize(entries[-1][ENTRY_MODE]) + entries[-1][ENTRY_STARTSECTOR]
    min,sec,frame = sector_2_msf(start_sector)
    entries.append([mode, min, sec, frame, start_sector, start_offset])

    return entries

def extract_track_data_nrg(raw_entries, bin_filename):
    " generates out of the raw data, we got from the nrg parser, useful data \
    we can use for our struct function"
    # first build an list we can calc with
  
    entries = []

    for i in range(0,len(raw_entries)):
        # track, track_raw, nexttrack, nexttrack_raw, start_sector, end_sector, sector_size, mode, start_offset, end_offset
        mode = get_mode_nrg(raw_entries[i][7], raw_entries[i][6])
        start_offset = raw_entries[i][8]
        if raw_entries[i][6] == 0: # Hack alert!
            start_sector = raw_entries[i][4]
        else:
            start_sector = start_offset / raw_entries[i][6]
        min, sec, frame = sector_2_msf(start_sector)

        entries.append([mode, min, sec, frame, start_sector, start_offset])

    # add the leadout
    mode = LEADOUT
    start_offset = raw_entries[-1][9]
    if raw_entries[-1][6] == 0: # Hack alert!
        start_sector = raw_entries[-1][9] / (raw_entries[-1][8] / raw_entries[i][4])
    else:
        start_sector = raw_entries[-1][9] / raw_entries[-1][6]
    min,sec,frame = sector_2_msf(start_sector)
    entries.append([mode, min, sec, frame, start_sector, start_offset])

    return entries

def extract_track_data_mds(raw_entries, bin_filename):
    " generates out of the raw data, we got from the mds parser, useful data \
    we can use for our struct function"
    # first build an list we can calc with
  
    entries = []

    for i in range(0,len(raw_entries)):
        # track, mode, sectsize, sector, sectors, offset
        mode = get_mode_mds(raw_entries[i][1], raw_entries[i][2])
        min, sec, frame = sector_2_msf(raw_entries[i][3])
        start_sector = raw_entries[i][3]
        start_offset = raw_entries[i][5]

        entries.append([mode, min, sec, frame, start_sector, start_offset])

    # add the leadout
    mode = LEADOUT
    start_offset = os.path.getsize(bin_filename)
    start_sector = (start_offset - entries[-1][ENTRY_STARTOFFSET]) / mode_2_sectorsize(entries[-1][ENTRY_MODE]) + entries[-1][ENTRY_STARTSECTOR]
    min,sec,frame = sector_2_msf(start_sector)
    entries.append([mode, min, sec, frame, start_sector, start_offset])

    return entries

def build_struct(entries, bin_file_handle, num_tracks, comment):
    "generates the c struct we need"
    # loop over the entries
    tracks = ""
    for entry in entries:
        tracks+=struct.pack(track_struct,*entry)

    return struct.pack(cdemu_struct,tracks,bin_file_handle,num_tracks,comment)

def get_device_from_drive_number(devicenumber):
    if not devicenumber.isdigit():
        raise CDEmu_Error, "Device number is not a digit"

    # we'll try /dev/cdemu0 and ... 
    try:
        device = device_path + devicenumber
        if stat.S_ISBLK(os.stat(device)[stat.ST_MODE]):
            return device
    except:
        try:
            # /dev/cdemu/0
            device = device_path + '/' + devicenumber
            if stat.S_ISBLK(os.stat(device)[stat.ST_MODE]):
                return device    
        except:
            pass        
    # its no blockdevice .. argh
    raise CDEmu_Error, "Do you have cdemu module loaded? - could't find block device. If you've no devfs/udev make sure you created them."

def load_cd(device_name, filename, comment):
    # read the CD image.
    filename = os.path.abspath(filename)

    cd_image_type = [ [ "ISO-9660 (ISO)", r'.+[.]iso$', parse_iso, extract_track_data_cue ],
                      [ "Alcohol 120% (MDS/MDF)", r'.+[.]mds$', parse_mds, extract_track_data_mds ],
                      [ "CDRWin (CUE/BIN)", r'.+[.]cue$', parse_cue, extract_track_data_cue ],
                      [ "CloneCD (CCD/IMG)", r'.+[.]ccd$', parse_ccd, extract_track_data_ccd ],
                      [ "Nero Burning ROM (NRG)", r'.+[.]nrg$', parse_nrg, extract_track_data_nrg ] ]

    for image_type in cd_image_type:
        image_regex = re.compile(image_type[1], re.IGNORECASE)
        if image_regex.search(filename):
            # Found supported CD image type, so parse the CD image file.
            if verbose: print "CD Image seems to be a '%s' file." % image_type[0]
            image_parser = image_type[2]
            supposable_bin_filename, raw_entries = image_parser(filename)
            # find the real bin-file
            bin_filename = find_bin(supposable_bin_filename, filename)
            if verbose: print "real (existing) CD image file: " , bin_filename
            # format the data into a way we need it
            image_track_data_extractor = image_type[3]
            tracks = image_track_data_extractor(raw_entries, bin_filename)
            if verbose: print "reformated tracks: ", tracks
            break
    else:
        # catch errors
        raise CDEmu_Error, "Unknown format of input file."

    # open the dev-device
    try:
        device = os.open(device_name, os.O_RDONLY | os.O_NONBLOCK)
        file = os.open(bin_filename,  os.O_RDONLY)
        # bild the c struct
        # -1 as we don't count the leadout
        c_struct = build_struct(tracks, file , len(tracks) - 1, comment)
  
        buf = array.array('c', c_struct);
        fcntl.ioctl(device, CDEMU_LOADCD, buf, 1)
        os.close(device)
        
    except OSError, msg:
        raise CDEmu_Error, msg

def unload_cd(device_name):
    try:
        # open the dev-device
        device = os.open(device_name, os.O_RDONLY | os.O_NONBLOCK)
        fcntl.ioctl(device, CDEMU_UNLOADCD, 0)
        os.close(device)
    except IOError, msg: raise CDEmu_Error, msg
    except OSError, msg: raise CDEmu_Error, msg

if __name__ == '__main__':
    "only for testing this lib ... no use in the real life"
    # we need excatly one parameter
    if len(sys.argv) != 2:
        print "we need one parameter .. and that should be an filename ;-)"
        sys.exit()

    #load_cd("/dev/cdemu/0", sys.argv[1],"blabla")
    #unload_cd("/dev/cdemu/0")
    #print build_struct(tracks, 0, len(tracks), "blabla")
